Format MS COFF

Public loaddll

Extrn kernel_alloc
Extrn kernel_free
Extrn PFileBuffer

Section ".text" Code Readable Writable Align 16

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; FROM: I40EMUL.INC
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

loaddll:
         push  ebx esi edi
; eax contains pointer to loaded file                           ;
         mov   ebx, eax                                         ;
; 1. Calculate image size & allocate memory for image           ;
         movzx ecx, word [eax + 2]                              ; NumberOfSections
         lea   edx, [eax + 20 + 16]                             ; pointer to SizeOfRawData
         xor   esi, esi                                         ;
@@:                                                             ;
         add   esi, [edx]                                       ;
         add   esi, 0xF                                         ; Alig-
         and   esi, not 0xF                                     ;     -nment
         add   edx, 40                                          ; SizeOf(SectionHeader)
         loop  @b                                               ;
         ;pushad                                                 ;
         ;call  .initialize_heap                                 ;
         ;mov   ecx, [esp + 4]                                   ;
         ;call  .allocate_heap                                   ;
         ;popad                                                  ; allocate_heap store result to [esp + 20h]
         push  esi
         call  kernel_alloc
         test  eax, eax                                         ; after popad result in eax
         jnz   @f                                               ;
; no memory in user space, user has already been notified       ;
         ;and   dword [esp + 20h], 0                             ; Returned Value
         xor   eax, eax
         pop   edi esi ebx
         ret                                                    ;
@@:                                                             ;
; eax = base addr in user space                                 ;
; ebx = pointer to loaded DLL data                              ;
; 2. Copy image data                                            ;
        mov    edi, eax                                         ;
        ;add    edi, [base]                                      ; it seems unnecessary in our case
        movzx  ecx, word [ebx + 2]                              ; NumberOfSections
        lea    edx, [ebx + 20 + 16]                             ; pointer to SizeOfRawData
@@:                                                             ;
        mov    [edx + 12 - 16], edi                             ;
        mov    esi, [edx + 20 - 16]                             ;
        test   esi, esi                                         ;
        jnz    .copy                                            ;
        add    edi, [edx + 16 - 16]                             ;
        jmp    .next                                            ;
.copy:                                                          ;
        add    esi, ebx                                         ;
        push   ecx                                              ;
        mov    ecx, [edx + 16 - 16]                             ;
        rep    movsb                                            ;
        pop    ecx                                              ;
.next:                                                          ;
        add    edi, 0xF                                         ;  Alig-
        and    edi, not 0xF                                     ;       -nment 
        add    edx, 40                                          ; SizeOf(SectionHeader)
        loop   @b                                               ;
; 3. Fixup COFF symbols                                         ;
; symbols table = ebx  +  [ebx + 8]                             ;
; strings ptr = (symbols table) + [ebx + 12] * 18               ;
        mov    ecx, [ebx + 12]                                  ; NumberOfSymbols
        mov    edx, ebx                                         ;
        add    edx, [ebx + 8]                                   ; PointerToSymbolTable
        lea    esi, [ecx + ecx * 8]                             ;
        lea    esi, [edx + esi * 2]                             ;
.fix_sym:                                                       ;
        movsx  edi, word [edx + 12]                             ;
        test   edi, edi                                         ;
        jnz    .internal                                        ;
; current implementation does not support exports               ;
; this is kernel limitation, not my                             ;
        and    dword [edx + 8], 0                               ;
        jmp    .next_sym                                        ;
.internal:                                                      ;
        js     .next_sym                                        ;
        dec    edi                                              ;
        shl    edi, 3                                           ;
        lea    edi, [edi + edi * 4]                             ;
        mov    edi, [ebx + 20 + edi + 12]                       ;
        ;sub    edi, [base]                                      ; it seems unnecessary in our case
        add    [edx + 8], edi                                   ;
.next_sym:                                                      ;
        add    edx, 18                                          ; SizeOf(Symbol)
        loop   .fix_sym                                         ;
; 4. Fixup COFF relocations                                     ;
        movzx  ecx, word [ebx + 2]                              ; NumberOfSections
        lea    esi, [ebx + 20]                                  ; pointer to SectionHeader
.fix_sec_reloc:                                                 ;
        mov    edi, [esi + 24]                                  ; PointerToRelocations
        add    edi, ebx                                         ;
        push   ecx                                              ;
        movzx  ecx, word [esi + 32]                             ; NumberOfRelocations
        jecxz  .next_sec_reloc                                  ;
.fix_reloc:                                                     ;
        mov    edx, [edi + 4]                                   ;
        lea    edx, [edx + edx * 8]                             ;
        lea    edx, [ebx + edx * 2]                             ;
        add    edx, [ebx + 8]                                   ;
        mov    edx, [edx + 8]                                   ;
        mov    eax, [edi]                                       ;
        add    eax, [esi + 12]                                  ; SectionHeader.VirtualAddress
        cmp    word [edi + 8], 6                                ;
        jz     .dir32                                           ;
        cmp    word [edi + 8], 20                               ;
        jz     .rel32                                           ;
        ;pushad                                                 ;
        ;push   30h                                             ;
        ;push   aWarning                                        ;
        ;push   aUnknownReloc                                   ;
        ;push   0                                               ;
        ;call   [MessageBoxA]                                   ;
        ;popad                                                  ;
        jmp    .next_reloc                                      ;
.rel32:                                                         ;
        sub    edx, eax                                         ;
        sub    edx, 4                                           ;
.dir32:                                                         ;
        add    [eax], edx                                       ;
.next_reloc:                                                    ;
        add    edi, 10                                          ;
        loop   .fix_reloc                                       ;
.next_sec_reloc:                                                ;
        add    esi, 40                                          ; SizeOf(SectionHeader)
        pop    ecx                                              ;
        loop   .fix_sec_reloc                                   ;
; 5. Lookup 'EXPORTS' symbol                                    ;
        mov    eax, ebx                                         ;
        add    eax, [ebx + 8]                                   ; PointerToSymbolTable
        mov    ecx, [ebx + 12]                                  ; NumberOfSymbols
.find_exp:                                                      ;
        cmp    dword [eax], 'EXPO'                              ;
        jnz    @f                                               ;
        cmp    dword [eax + 4], 'RTS'                           ;
        jz     .found_exp                                       ;
@@:                                                             ;
        add    eax, 18                                          ;
        loop   .find_exp                                        ;
; 5b. Lookup '_EXPORTS' symbol                                  ;
        mov    eax, ebx                                         ;
        add    eax, [ebx + 8]                                   ; PointerToSymbolTable
        mov    ecx, [ebx + 12]                                  ; NumberOfSymbols 
.find__exp:                                                     ;
        cmp    dword [eax], '_EXP'                              ;
        jnz    @f                                               ;
        cmp    dword [eax + 4], 'ORTS'                          ;
        jz     .found_exp                                       ;
@@:                                                             ;
        add    eax, 18                                          ;
        loop   .find__exp                                       ;
        ;push   30h                                             ;
        ;push   aWarning                                        ;
        ;push   aExportsNotFound                                ;
        ;push   0                                               ;
        ;call   [MessageBoxA]                                   ;
; FIXME: if we here then EXPORTS was not found 
;       need to use kernel_free here to free memory allocated by kernel_alloc        
        xor    eax, eax                                         ;
        jmp    @f                                               ;
.found_exp:                                                     ;
        mov    eax, [eax + 8]                                   ;
@@:                                                             ;
        ;mov    [esp + 20h], eax                                 ; Returned Value
        ;push   ebx                                             ;
        ;call   free_big                                        ;
        pop   edi esi ebx
        ret                                                     ;
